@@ -27,13 +27,13 @@ const DEFAULT_INITIAL_MAX_STREAMS_UNI: u64 = 100;
 /// `UdpSocket` should only be used directly if the default configuration used
 /// by `QuicListener::bind` does not meet the required use case.
 pub struct QuicSocket {
-    inner: tokio::net::UdpSocket,
-    addr: SocketAddr,
+    pub inner: tokio::net::UdpSocket,
+    pub addr: SocketAddr,
 }
 
 pub struct QuicListener {
-    socket: QuicSocket,
-    connection: std::pin::Pin<std::boxed::Box<quiche::Connection>>,
+    pub socket: QuicSocket,
+    pub connection: std::pin::Pin<std::boxed::Box<quiche::Connection>>,
 }
 
 impl QuicSocket {
@@ -66,6 +66,21 @@ impl QuicSocket {
         }
     }
 
+    pub async fn recv_from(&self) -> Result<(usize, SocketAddr), io::Error> {
+        let mut buf = [0; 65535];
+        loop {
+            match self.inner.recv_from(&mut buf).await {
+                Ok((len, from)) => return Ok((len, from)),
+                Err(_) => {
+                    return Err(io::Error::new(
+                        io::ErrorKind::WouldBlock,
+                        "recv would block",
+                    ))
+                }
+            };
+        }
+    }
+
     /// Creates new `QuicSocket` from a previously bound `std::net::UdpSocket`.
     ///
     /// The conversion assumes nothing about the underlying socket; it is left up to the user to set it in
@@ -262,13 +277,7 @@ impl QuicListener {
     /// }
     /// ```
     pub async fn send(&mut self, out: &mut [u8]) -> Result<usize, io::Error> {
-        let (write, info) = match self.connection.send(out) {
-            Ok(v) => v,
-            Err(quiche::Error::Done) => {
-                return Err(io::Error::new(io::ErrorKind::Other, "done writing"))
-            }
-            Err(_) => return Err(io::Error::new(io::ErrorKind::Other, "could not send")),
-        };
+        let (write, info) = self.connection.send(out).unwrap();
         self.send_to(&mut out[..write], &info).await
     }
 
